/* NEC Vr4155 CSI driver
 * CSI: Clocked Serial Interface (SPI)
 *
 * (C) Copyright 2007 Media Lab. <linux@mlb.co.jp>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 *
 */

#include <config.h>

#if defined(CONFIG_SPI) && defined(CONFIG_VR4155_CSI)
/*#define DEBUG	1*/
#include <common.h>
#include <malloc.h>
#include <net.h>
#include <command.h>
#include <asm/io.h>
#include <asm/cacheops.h>
#include <spi.h>

#define CSI_BASE	0xAF009000
#define CSI_MODE	0xAF009000 /* R/W CSI Mode Control register                 */
#define  CSI_MODE_CSOT		(1)
#define  CSI_MODE_CCL		(1<<5)
#define  CSI_MODE_TRMD		(1<<6)
#define  CSI_MODE_CSIE		(1<<7)
#define  CSI_MODE_WAIT(n)	((n)<<16)
#define CSI_CLKSEL	0xAF009004 /* R/W CSI Clock Select register                 */
#define  CSI_CLKSEL_DAP		(1<<3)
#define  CSI_CLKSEL_CKP		(1<<4)
#define  CSI_CLKSEL_16M		(0)
#define  CSI_CLKSEL_4M		(1)
#define  CSI_CLKSEL_2M		(2)
#define  CSI_CLKSEL_1M		(3)
#define  CSI_CLKSEL_521		(4)
#define  CSI_CLKSEL_260		(5)
#define  CSI_CLKSEL_130		(6)
#define  CSI_CLKSEL_SLAVE	(7)
#define CSI_CNT		0xAF009008 /* R/W CSI Control register                      */
#define  CSI_CNT_CSIRST		(1<<28)
#define  CSI_CNT_T_TRGEN	(1<<27)
#define  CSI_CNT_T_FIFOF	(1<<26)
#define  CSI_CNT_R_TRGEN	(1<<19)
#define  CSI_CNT_R_FIFOF	(1<<18)
#define  CSI_CNT_UNDER_E	(1<<13)
#define  CSI_CNT_OVERF_E	(1<<12)
#define  CSI_CNT_CSIEND_E	(1<<8)
#define  CSI_CNT_T_TRGR_E	(1<<4)
#define  CSI_CNT_R_TRGR_E	(1<<1)
#define CSI_INT		0xAF00900C /* R/W CSI Interrupt Status register             */
#define  CSI_INT_UNDER		(1<<13)
#define  CSI_INT_OVERF		(1<<12)
#define  CSI_INT_CSIEND		(1<<8)
#define  CSI_INT_T_TRGR		(1<<4)
#define  CSI_INT_R_TRGR		(1)
#define CSI_IFIFOL	0xAF009010 /* R/W CSI Receive FIFO Level Indicate register  */
#define CSI_OFIFOL	0xAF009014 /* R/W CSI Transmit FIFO Level Indicate register */
#define CSI_IFIFO	0xAF009018 /*  R  CSI Receive FIFO window register          */
#define CSI_OFIFO	0xAF00901C /*  W  CSI Transmit FIFO window register         */
#define CSI_FIFOTRG	0xAF009020 /* R/W CSI FIFO Trigger Level register           */
#define  CSI_FIFOTRG_T_TRG(n)	((n)<<8)
#define  CSI_FIFOTRG_R_TRG(n)	(n)

#define CMU_CLK_MSK	0xAF000060 /* clock mask */
#define CMU_CLK_MSK_CSI	(1<<5)

#define SCUPINMUX3REG	0xAF008008
#define SCUPINMUX3_CSI_MASK	(0x00FC0000)
#define SCUPINMUX3_CSI	  	(0)
/*
#define DATA_LENGTH_16		1
*/

#define csi_outl( reg, val ) (*((volatile unsigned int *)(reg)) = (val) )
#define csi_inl( reg ) (*((volatile unsigned int *)(reg)))

void spi_init (void)
{
	debug ( "%s\n", __FUNCTION__ );
	*((volatile unsigned int *)(CMU_CLK_MSK)) |= CMU_CLK_MSK_CSI;
	*((volatile unsigned int *)(SCUPINMUX3REG)) &= ~SCUPINMUX3_CSI_MASK;
}

/* called from cmd_spi.c and rtc/ds1306.c */
int spi_xfer(spi_chipsel_type chipsel, int bitlen, uchar *out, uchar *in)
{
	int len, i;
#ifdef DATA_LENGTH_16
	ushort *dout = (ushort *)out;
	ushort *din  = (ushort *)in;
#else
	uchar *dout = out;
	uchar *din  = in;
#endif

	debug ( "%s(%p,%d,%p,%p)\n", __FUNCTION__, chipsel, bitlen, dout, din );
	csi_outl( CSI_CNT, 0 ); /* deassert CSI reset */
#ifdef DATA_LENGTH_16
	csi_outl( CSI_MODE, CSI_MODE_WAIT(4) | CSI_MODE_TRMD | CSI_MODE_CCL );
#else
	csi_outl( CSI_MODE, CSI_MODE_WAIT(4) | CSI_MODE_TRMD ); /* 1 clock wait, 8bit, MSB first */
#endif
	csi_outl( CSI_CLKSEL, CSI_CLKSEL_260 ); 
	csi_outl( CSI_INT, 0x0000F111 );
#ifdef DATA_LENGTH_16
	csi_outl( CSI_FIFOTRG, CSI_FIFOTRG_T_TRG(4) );
#else
	csi_outl( CSI_FIFOTRG, CSI_FIFOTRG_T_TRG(3) );
#endif
	csi_outl( CSI_CNT,    CSI_CNT_T_TRGEN|CSI_CNT_T_TRGR_E|CSI_CNT_CSIEND_E );

	/* flush FIFO */
	csi_outl( CSI_OFIFOL, 0x0 );
	csi_outl( CSI_IFIFOL, 0x0 );

#ifdef DATA_LENGTH_16
	len = (bitlen + 15) / 16;
#else
	len = (bitlen + 7) / 8;
#endif

#ifdef DEBUG
	debug ( "out(%d):", len );
	for ( i = 0; i < len; i++ )
	  debug ( "%02X ", dout[i] );
	debug ( "\n" );
#endif	

	/* fill output FIFO */
	i = 0;
	while ( len-- ) {
	  csi_outl( CSI_OFIFO, *dout );
	  dout++;
#ifdef DATA_LENGTH_16
	  if ( ++i >= 16 )
	    break;
#else
	  if ( ++i >= 32 )
	    break;
#endif
	}

	if(chipsel != NULL) {
		(*chipsel)(1);	/* select the target chip */
	}
	/* start transemit */
	csi_outl( CSI_MODE, (csi_inl( CSI_MODE ) | CSI_MODE_CSIE) );

	while ( len > 0 ) {
	  while ( !(csi_inl( CSI_INT ) & CSI_INT_T_TRGR) )
	    udelay(15);

	  while( csi_inl( CSI_IFIFOL ) ) {
	    *(din++) = csi_inl( CSI_IFIFO );
	  }

	  i = 0;
	  while( len-- ) {
	    csi_outl( CSI_OFIFO, *dout );
	    dout++;
	    if ( ++i >= 16 )
	      break;
	  } 
	  csi_outl( CSI_INT, CSI_INT_T_TRGR );
	}

	/* wait for finish */
	while( !(csi_inl( CSI_INT) & CSI_INT_CSIEND) || csi_inl( CSI_MODE ) & CSI_MODE_CSOT );
	  udelay(15);

	/* stop */
	csi_outl( CSI_MODE, (csi_inl( CSI_MODE ) & ~CSI_MODE_CSIE) );

	debug ( "%08X=%08X\n", CSI_MODE, csi_inl(CSI_MODE) );
	debug ( "%08X=%08X\n", CSI_INT, csi_inl(CSI_INT) );
	debug ( "%08X=%08X\n", CSI_IFIFOL, csi_inl(CSI_IFIFOL) );
	debug ( "%08X=%08X\n", CSI_OFIFOL, csi_inl(CSI_OFIFOL) );
	while( csi_inl( CSI_IFIFOL ) ) {
	  *(din++) = csi_inl( CSI_IFIFO );
	}

	if(chipsel != NULL) {
		(*chipsel)(0);	/* deselect the target chip */
	}

#ifdef DEBUG
	debug ( "in(%d):", ((uchar*)din) - in );
	for ( i = 0; &in[i] < (uchar*)din; i++ )
	  debug ( "%02X ", in[i] );
	debug ( "\n" );
#endif	
	return(0);
}

#endif /* defined(CONFIG_SPI) && defined(CONFIG_VR4155_CSI) */
