/*
 * rx5c348.c
 *
 * RTC driver for Richo rx5c348A/B
 * Copyright (C) 2007 Media Lab.  ito@mlb.co.jp 
 *
 * Based on rtc/rs5c372.c -
 * Copyright (C) 2004 Gary Jennejohn garyj@denx.de
 *
 * Based in part in ds1307.c -
 * (C) Copyright 2001, 2002, 2003
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 * Keith Outwater, keith_outwater@mvis.com`
 * Steven Scholz, steven.scholz@imc-berlin.de
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <command.h>
#include <rtc.h>
#include <spi.h>

#if defined(CONFIG_RTC_RX5C348) && (CONFIG_COMMANDS & CFG_CMD_DATE) 

#define WRAPSIZE	0
/*
 * address 0 - 6 need to wait 31us 
 * It is necessary to wait 31uS to access address 0-6 after chip activated.
 * If you are using more than 260kHz, you can adjust with WRAPSIZE.
 * WRAPSIZE 1 means 8bit data will be inserted before reading adress 0.
 *
 */

#define DEBUG 0

#if DEBUG
static unsigned int rtc_debug = DEBUG;
#else
#define rtc_debug 0	/* gcc will remove all the debug code for us */
#endif

#define BUF_SIZE 	0x11
#define RATE_32768HZ	0x00	/* Rate Select 32.768KHz */

#define STATUS_XPT  0x10    /* data invalid because voltage was 0 */

#define USE_24HOUR_MODE 0x20
#define TWELVE_HOUR_MODE(n) ((((n) >> 5) & 1) == 0)
#define HOURS_AP(n)	(((n) >> 5) & 1)
#define HOURS_12(n)	bcd2bin((n) & 0x1F)
#define HOURS_24(n)	bcd2bin((n) & 0x3F)
#define MONTH(n)	bcd2bin((n) & 0x1F)
#define CENTURY_20	(0x80)
#define CENTURY_BIT(n)	((n) & CENTURY_20)


static uchar bin2bcd (unsigned int n);
static unsigned bcd2bin (uchar c);
#define WRAPPOS(a)	((((a)+WRAPSIZE)&0xF)+1)
#define REG_SEC		WRAPPOS(0x0)
#define REG_MIN		WRAPPOS(0x1)
#define REG_HOUR	WRAPPOS(0x2)
#define REG_WDAY	WRAPPOS(0x3)
#define REG_DAY		WRAPPOS(0x4)
#define REG_MONTH	WRAPPOS(0x5)
#define REG_YEAR	WRAPPOS(0x6)
#define REG_ADJUST	WRAPPOS(0x7)
#define REG_ALARMW_MIN	WRAPPOS(0x8)
#define REG_ALARMW_HOUR	WRAPPOS(0x9)
#define REG_ALARMW_WDAY	WRAPPOS(0xA)
#define REG_ALARMD_MIN	WRAPPOS(0xB)
#define REG_ALARMD_HOUR	WRAPPOS(0xC)
#define REG_CONTOL1	WRAPPOS(0xE)
#define REG_CONTOL2	WRAPPOS(0xF)
#define RTC_WRITE_BYTE	0x8
#define RTC_WRITE_BURST	0x0
#define RTC_READ_BYTE	0xC
#define RTC_READ_BURST	0x4

static int setup_done = 0;

static int rtc_read( uchar *buffer, int len )
{
	buffer[0] = (((0x10 - WRAPSIZE) << 4)&0xF0)|RTC_READ_BURST; /* burst read */
	return spi_xfer (spi_chipsel[CFG_SPI_RTC_DEVID], len*8, buffer, buffer);
}

static int rtc_write( uchar *buffer, int len )
{
	buffer[0] = ((0x10 - WRAPSIZE) << 4)|RTC_WRITE_BURST; /* burst write */
	return spi_xfer (spi_chipsel[CFG_SPI_RTC_DEVID], len*8, buffer, buffer);
}

static int
rx5c348_readram(unsigned char *buf, int len)
{
	int ret;

	ret = rtc_read(buf, len);

	if (ret != 0) {
		printf("%s: failed to read\n", __FUNCTION__);
		return ret;
	}

	if (buf[REG_CONTOL2] & STATUS_XPT)
		printf("### Warning: RTC lost power\n");

	return ret;
}

static int
rx5c348_enable(char *buf)
{
	int ret;

	ret = rx5c348_readram(buf, BUF_SIZE);
	if (ret != 0) {
		printf("%s: failed\n", __FUNCTION__);
		return ret;
	}
	if (buf[REG_CONTOL2] & STATUS_XPT) {
	  rtc_reset();
	  ret = rx5c348_readram(buf, BUF_SIZE);
	}
	setup_done = 1;

	return 0;
}

static void
rx5c348_convert_to_time(struct rtc_time *dt, unsigned char *buf)
{
	/* buf[0] is register 15 */
	dt->tm_sec = bcd2bin(buf[REG_SEC]);
	dt->tm_min = bcd2bin(buf[REG_MIN]);

	if (TWELVE_HOUR_MODE(buf[REG_CONTOL1])) {
		dt->tm_hour = HOURS_12(buf[REG_HOUR]) % 12;
		if (HOURS_AP(buf[REG_HOUR])) /* PM */
			dt->tm_hour += 12;
	} else /* 24-hour-mode */
		dt->tm_hour = HOURS_24(buf[REG_HOUR]);

	dt->tm_mday = bcd2bin(buf[REG_DAY]);
	dt->tm_mon = MONTH(buf[REG_MONTH]);
	dt->tm_year = bcd2bin(buf[REG_YEAR]);
	if ( CENTURY_BIT(buf[REG_MONTH]) ) {
	  dt->tm_year += 2000;
	}else{
	  dt->tm_year += 1900;
	}

	/* 0 is Sunday */
	dt->tm_wday = bcd2bin(buf[REG_WDAY] & 0x07);
	dt->tm_yday = 0;
	dt->tm_isdst= 0;

	if(rtc_debug > 2) {
		printf("rx5c348_convert_to_time: year = %d\n", dt->tm_year);
		printf("rx5c348_convert_to_time: mon  = %d\n", dt->tm_mon);
		printf("rx5c348_convert_to_time: mday = %d\n", dt->tm_mday);
		printf("rx5c348_convert_to_time: hour = %d\n", dt->tm_hour);
		printf("rx5c348_convert_to_time: min  = %d\n", dt->tm_min);
		printf("rx5c348_convert_to_time: sec  = %d\n", dt->tm_sec);
	}
}

/*
 * Get the current time from the RTC
 */
void
rtc_get (struct rtc_time *tmp)
{
	unsigned char buf[BUF_SIZE];

	rx5c348_enable( buf );

	if (!setup_done)
	  return;

	rx5c348_convert_to_time(tmp, buf);

	return;
}

/*
 * Set the RTC
 */
void
rtc_set (struct rtc_time *tmp)
{
	unsigned char buf[BUF_SIZE];
	int ret;

	rx5c348_enable(buf);

	if (!setup_done)
		return;

	if(rtc_debug > 2) {
		printf("rtc_set: tm_year = %d\n", tmp->tm_year);
		printf("rtc_set: tm_mon	 = %d\n", tmp->tm_mon);
		printf("rtc_set: tm_mday = %d\n", tmp->tm_mday);
		printf("rtc_set: tm_hour = %d\n", tmp->tm_hour);
		printf("rtc_set: tm_min	 = %d\n", tmp->tm_min);
		printf("rtc_set: tm_sec	 = %d\n", tmp->tm_sec);
	}

	buf[REG_SEC] = bin2bcd(tmp->tm_sec);
	buf[REG_MIN] = bin2bcd(tmp->tm_min);
	/* need to handle 12 hour mode */
	if ( TWELVE_HOUR_MODE(buf[REG_CONTOL1]) ) {
	  if (tmp->tm_hour < 12 )
	    buf[REG_HOUR] = bin2bcd(tmp->tm_hour);
	  else
	    buf[REG_HOUR] = bin2bcd(tmp->tm_hour - 12);
	  /* hour is 01 to 12 */
	  if ( buf[REG_HOUR] == 0 ) {
	    buf[REG_HOUR] = bin2bcd(12);
	  }
	} else {
	  buf[REG_HOUR] = bin2bcd(tmp->tm_hour);
	}
	
	buf[REG_WDAY] = bin2bcd(tmp->tm_wday);
	buf[REG_DAY] = bin2bcd(tmp->tm_mday);
	buf[REG_MONTH] = bin2bcd(tmp->tm_mon);
	if ( tmp->tm_year > 2000 ) {
	  buf[REG_MONTH] |= CENTURY_20;
	}
	buf[REG_YEAR] = bin2bcd(tmp->tm_year % 100);
	
	ret = rtc_write(buf, BUF_SIZE);
	if (ret != 0)
	  printf("rx5c348_set_datetime(), i2c_master_send() returned %d\n",ret);

	return;
}

/*
 * Reset the RTC. We set the date back to 1970-01-01.
 */
void
rtc_reset (void)
{
	unsigned char buf[BUF_SIZE];

	/* initialize with valid value */
	memset( buf, 0, sizeof(buf) );
	/* Jan. 1, 1970 was a Thursday */
	buf[REG_WDAY] = 4;
	buf[REG_DAY] = 1;
	buf[REG_MONTH] = 1;
	buf[REG_YEAR] = bin2bcd(70);
	buf[REG_CONTOL1] = USE_24HOUR_MODE;
	rtc_write( buf, BUF_SIZE );
	return;
}

static unsigned int
bcd2bin (unsigned char n)
{
	return ((((n >> 4) & 0x0F) * 10) + (n & 0x0F));
}

static unsigned char
bin2bcd (unsigned int n)
{
	return (((n / 10) << 4) | (n % 10));
}
#endif /* defined(CONFIG_RTC_RX5C348) && (CONFIG_COMMANDS & CFG_CMD_DATE) */
